#include "vad_test.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

#include "esp_system.h"
#include "esp_event.h"
#include "esp_event_loop.h"
#include "nvs_flash.h"

#include "esp_event.h"
#include "esp_event_loop.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "opus.h"
#include "debug.h"
#include "opus_types.h"
#include <silk/structs.h>
#include "opus_multistream.h"
#include "main.h"
#include "esp_log.h"
#include "API.h"
#include "structs_FIX.h"

#include "driver/i2s.h"
#include "driver/dac.h"

#define VAR_ARRAYS	1

#define SAMPLE_RATE			48000
#define BITRATE				24000
#define SAMPLE_TIME_MS		20
#define SAMPLE_BITS			32    
#define SAMPLE_T			int32_t 
#define SAMPLES_SHORT		(SAMPLE_TIME_MS * SAMPLE_RATE / 1000)			// 960 ???? 
#define SAMPLES_LEQ			(SAMPLE_RATE * LEQ_PERIOD)

#define DMA_BUFF_LEN		512				// (SAMPLES_SHORT)/8
#define DMA_BANKS			32							// 32

#define MIC_BITS			24													// valid number of bits in I2S data
#define MIC_CONVERT(s)		(((s >> (SAMPLE_BITS - MIC_BITS))>>5) & 0xFFFF)				// shift it right to justify it.  and for the moment turn it into 16 bit by masking out .

#define I2S_PORT			I2S_NUM_0

#define I2S_TASK_PRI		4
#define I2S_TASK_STACK		2048

#define CHANNELS			1

#define FRAME_SIZE SAMPLES_SHORT
#define MAX_FRAME_SIZE 6*SAMPLES_SHORT
#define MAX_PACKET_SIZE (3*1276)

// #define CAPTURE_ZEROS		 

#define OVERRIDE_silk_noise_shape_analysis_FIX	1

// #define HAVE_MALLOC

static const char *TAG = "IHM-CORONA-VAD";

// Static buffer for block of samples
static int32_t samples[SAMPLES_SHORT] __attribute__((aligned(4)));


unsigned char cbits[MAX_PACKET_SIZE];


silk_VAD_state VAD_STATE;

OpusEncoder * Oencoder = NULL;

// Data we push to 'samples_queue'
struct sum_queue_t {
	// Sum of squares of mic samples, after Equalizer filter
	float sum_sqr_SPL;
	// Sum of squares of weighted mic samples
	float sum_sqr_weighted;
	// Debug only, FreeRTOS ticks we spent processing the I2S data
	uint32_t proc_ticks;
};

//
// I2S Microphone sampling setup 
//
// Simple I2S read hardware setup
//
//
void mic_i2s_init() {

	static const i2s_config_t i2s_conf = {
		.mode = I2S_MODE_MASTER | I2S_MODE_RX /*| I2S_MODE_DAC_BUILT_IN  | I2S_MODE_PDM*/,
		.sample_rate = SAMPLE_RATE,
		.bits_per_sample = SAMPLE_BITS,
		.channel_format = I2S_CHANNEL_FMT_ONLY_RIGHT,
		.communication_format = I2S_COMM_FORMAT_I2S | I2S_COMM_FORMAT_I2S_MSB /*| I2S_COMM_FORMAT_PCM_LONG*/,
		.intr_alloc_flags = I2S_TASK_PRI,
		.dma_buf_count = DMA_BANKS,
		.dma_buf_len = DMA_BUFF_LEN,
		.use_apll = true,
		.tx_desc_auto_clear = false,
		.fixed_mclk = 0
		};  
	
	static const i2s_pin_config_t pin_config = {
    	.bck_io_num = 0,
		.ws_io_num = 5,
		.data_in_num = 26,
		.data_out_num = I2S_PIN_NO_CHANGE
		};
	
	i2s_driver_install(I2S_PORT, &i2s_conf, 0, NULL);
	i2s_set_pin(I2S_PORT, &pin_config);
}

//
// This becomes an Rtos task to read up the I2S stream of data an package it into sample packets for the main loop to process
//
// To investigate - do Opus / silk encoding here.
// 
void mic_i2s_reader_task(QueueHandle_t mic_queue) {
	
	int nbBytes = 0;	
	size_t bytes_read = 0;	

	// Discard first blocks ( until mic reads proper data ), microphone has startup time
	int test = -1;
	while (test < 0) {
		i2s_read(I2S_PORT, &samples, SAMPLES_SHORT*4, &bytes_read, portMAX_DELAY);
		for (int i = 0; i < SAMPLES_SHORT; i++) {
			if (samples[i] > 10)
				test = 1;
			break;
			}
	    }

	while (true) {
		esp_err_t e = i2s_read(I2S_PORT, (void*)&samples, SAMPLES_SHORT * sizeof(opus_int32), &bytes_read, portMAX_DELAY);
		if (e == ESP_OK){
//			printf("send  %d\n", bytes_read);
 			xQueueSend(mic_queue, &bytes_read, portMAX_DELAY);  
			}
		}
}



//
//
//
//
esp_err_t event_handler(void *ctx, system_event_t *event)
{
	return ESP_OK;
}

void app_main(void)
{
	
	int err=0;
	int counter = 0;	

	size_t count = 0;
	int readBytes = 0;	
	
	size_t bytes_read = 0;		
	
	opus_int16 q[SAMPLES_SHORT] __attribute__((aligned(2)));

	// Create FreeRTOS queue	
 	ESP_LOGI(TAG, "[ 1 ]Create Task Queue");	
 	
	QueueHandle_t mic_event_queue = xQueueCreate(8, sizeof(opus_int));	

	ESP_LOGI(TAG, "[ 2 ]Create Opus encoder");	
 	Oencoder = opus_encoder_create(SAMPLE_RATE, CHANNELS, OPUS_APPLICATION_VOIP, &err);
	
//	opus_encoder_init(Oencoder);
	
	if (err == ESP_OK)
	{
		ESP_LOGI(TAG, "Created Opus encoder");	
	
	//	err = opus_encoder_ctl(Oencoder, OPUS_SET_BITRATE(BITRATE));
	// 	err = opus_encoder_ctl(Oencoder, OPUS_SET_COMPLEXITY(5));						// Default is 10
																						// Default bandwidth is 20khz
		// Set DTX so that encoder uses silk and thus VAD
									
	 	err = opus_encoder_ctl(Oencoder, OPUS_SET_DTX(1));
		err = opus_encoder_ctl(Oencoder, OPUS_SET_PACKET_LOSS_PERC(100));				// Allow up to 100% packet loss
	
		ESP_LOGI(TAG, "[ 3 ] Initialize MIC I2S");
		
		mic_i2s_init();																	// Set up the mic interface
	
		// Fire up the mic reader task and pass it a pointer to the queue.
		ESP_LOGI(TAG, "[ 4 ] Initialize MIC I2S Stream task");
		xTaskCreate(&mic_i2s_reader_task, "MIC I2S Sampler", I2S_TASK_STACK, (void *) mic_event_queue, 6, NULL);	

		//
		// Catch a bunch of samples from the microphone
		// Pass to encoder and make a decision on voice
		// Also pass on to report process ( to be written ) 
		ESP_LOGI(TAG, "[ 4 ] Enter Processing loop");
		while (1) {
			// This should be blocking
			// All the send does is tell us if it was a good or bad capture.
			xQueueReceive(mic_event_queue, &count, portMAX_DELAY);
			int len = count / sizeof(opus_int32);
		
//			printf("Process %d \n", len);
			
			if (count > 0) {
//				ESP_LOGI(TAG, "process %d bytes\n", count);
				for(int i = 0 ; i < len ; i++)
				{
#ifdef CAPTURE_ZEROS
					q[i] = 0;				  
#else
					q[i] = MIC_CONVERT(samples[i]);
	//				printf("%d %d %d\n", i, samples[i], q[i]);
					printf("%d\n", q[i]);
#endif // DEBUG	
				}

//				counter++;
//				ESP_LOGI(TAG, "Launch encode\n");
			
//	 			readBytes = opus_encode(Oencoder, q, len, cbits, MAX_PACKET_SIZE);
		
/*				if(readBytes < 0) {
					ESP_LOGI(TAG, "encode failed: %s\n", opus_strerror(readBytes));
				}    
				else {
					ESP_LOGI(TAG, "encoded: %d\n", counter);	
					}  */
				}
			}
		}
	ESP_LOGI(TAG, "Exiting main!");	
}